class Settings:
    MAX_ANSWER_LENGTH = 2000
    MAX_EXPRESSION_CONSTRAINT_LENGTH = 1000
    MAX_NUM_OPTIONS = 20
    MIN_NUM_OPTIONS = 2
    MAX_OPTION_LENGTH = 1000
    MAX_QUESTION_LENGTH = 100000
